module Modular where

import Data.List
import Term
import Rule
import TRS
import Result
import qualified Proof as Proof

type Decomposition = TRS -> [(TRS,TRS)]

proof :: String -> String -> TRS -> (TRS,TRS) -> String -> String -> String
proof method desc trs (r,s) rProof sProof =
  Proof.splitproof
    method
    trs
    (r,s)
    desc
    (rProof ++ "\n" ++ sProof)

run :: ((TRS,TRS) -> String -> String -> String) -> Criterion -> [(TRS,TRS)] -> IO Result
run _ _ [] = return MAYBE
run p thm ((r,s):products) = do
  m1 <- thm r
  case m1 of
    YES rproof -> do
      m2 <- thm s
      case m2 of
        YES sproof -> return (YES (p (r,s) rproof sproof))
        _ -> run p thm products
    _ -> run p thm products

intersecting :: Eq a => [a] -> [a] -> Bool
intersecting xs ys = intersect xs ys /= []

grouping :: Eq a => ([a] -> a -> Bool) -> [a] -> [a] -> ([a], [a])
grouping p acc xs
  | acc' == [] = (acc, xs)
  | otherwise  = grouping  p (acc ++ acc') (xs \\ acc')
  where
    acc' = [x | x <- xs, p acc x]

symEq :: Eq a => ([a],[a]) -> ([a],[a]) -> Bool
symEq (xs1,ys1) (xs2,ys2) = 
  seteq xs1 xs2 && seteq ys1 ys2
  ||
  seteq xs1 ys2 && seteq ys1 xs2
  where
    seteq a b = subset a b && subset b a


-- Toyama 1987
decomposeDisjointUnion :: Criterion -> Criterion
decomposeDisjointUnion thm trs =
  run
    (proof
      "Disjoint Union (Toyama 1987)"
      "Signatures of both TRSs are disjoint. "
      trs)
    thm
    (disjointUnion trs)

disjointUnion :: Decomposition
disjointUnion [] = []
disjointUnion (rule:trs) =
    let (acc,trs') = grouping p [rule] trs
        products = disjointUnion trs' in
    nubBy
      symEq
      ([(acc,trs')] ++
      [(acc++xs,ys) | (xs,ys) <- products, ys /= []] ++
      [(xs,acc++ys) | (xs,ys) <- products, xs /= []])
  where
    p acc rl = intersecting (TRS.functions acc) (Rule.functions rl)

-- Ohlebusch 1994
decomposeCS :: Criterion -> Criterion
decomposeCS thm trs
  | not (TRS.leftLinear trs) = return MAYBE
  | otherwise =
    run
      (proof
        "Constructor Sharing (Toyama 1987)"
        "Signatures of both TRSs only share constructors. "
        trs)
      thm
      (constructorSharing trs)

isConstructorSharing :: TRS -> TRS -> Bool
isConstructorSharing r1 r2 =
  not (intersecting (TRS.definedSymbols r1) (TRS.definedSymbols r2))

constructorSharing :: Decomposition
constructorSharing [] = []
constructorSharing (rule:trs) =
    let (acc,trs') = grouping p [rule] trs
        products = constructorSharing trs' in
    nubBy
      symEq
      ([(acc,trs')] ++
      [(acc++xs,ys) | (xs,ys) <- products, ys /= []] ++
      [(xs,acc++ys) | (xs,ys) <- products, xs /= []])
  where
    p acc rl = not (isConstructorSharing acc [rl])

-- Ohlebusch 1994
-- NOTE: composablity is not sufficient condition of conluence 
-- but it is necessary condition.
isComposable :: TRS -> TRS -> Bool
isComposable r1 r2 =
  not (intersecting ds1 (TRS.constructors r2))
  &&
  not (intersecting ds2 (TRS.constructors r1))
  &&
  subset [lr | lr@(F f _,_) <- r1++r2, elem f ds] (intersect r1 r2)
  where
    ds1 = TRS.definedSymbols r1
    ds2 = TRS.definedSymbols r2
    ds  = intersect ds1 ds2

power :: [a] -> [[a]]
power [] = [ [] ]
power (x : xs) = yss ++ [ x : ys | ys <- yss ]
  where yss = power xs

composable :: Decomposition
composable trs =
  nubBy
    symEq
    [(r1,r2) |
      r1 <- power trs,
      r2 <- power trs,
      -- cut off
      not (subset r1 r2) && not (subset r2 r1),
      subset trs (r1 ++ r2),
      isComposable r1 r2]

-- Ohlebusch 1994
decomposeLP :: Criterion -> Criterion
decomposeLP thm trs =
  run
    (proof
      "Layer-Preserving Decomposition (Ohlebusch 1994)"
      "Two systems are composable and layer-preserving. "
      trs)
    thm
    (layerPreserving trs)

isLayerPreserving :: TRS -> TRS -> Bool
isLayerPreserving r1 r2 =
  all
    (\(_, r) -> rootPreserve r (TRS.functions r1 \\ TRS.functions r2))
    (r1 \\ r2)
  where
    rootPreserve (F f _) fs = elem f fs
    rootPreserve (V _) _ = False

layerPreserving :: Decomposition
layerPreserving trs =
  nub
    [(r1,r2) | (r1,r2) <- composable trs, isLayerPreserving r1 r2]

-- Ohlebusch 1994
-- NOTE: composablity is not sufficient condition of conluence 
-- but it is necessary condition.
decomposeHC :: Criterion -> Criterion
decomposeHC thm trs
  | not (TRS.leftLinear trs) = return MAYBE
  | otherwise =
    run
      (proof
        "Hierarchical Combination (Ohlebusch 1994)"
        "NOT SUFFICIENT CONDITION! "
        trs)
      thm
      (hierarchicalCombination trs)

isHierarchicalCombination :: TRS -> TRS -> Bool
isHierarchicalCombination extension base =
  not (intersecting fs ds)
  where
    fs = TRS.functions base
    ds = TRS.definedSymbols extension

hierarchicalCombination :: Decomposition
hierarchicalCombination trs =
  [(r1,r2) |
    r1 <- power trs,
    r2 <- power trs,
    -- cut off
    not (subset r1 r2) && not (subset r2 r1),
    subset trs (r1 ++ r2),
    isHierarchicalCombination r1 r2]

-- Composable Hierarchy
groupByRoot :: TRS -> [TRS]
groupByRoot [] = []
groupByRoot (rule:trs) =
  (rule:xs) : groupByRoot (trs \\ xs)
  where
    xs = [rule' |
         rule' <- trs,
         intersecting (TRS.definedSymbols [rule]) (TRS.definedSymbols [rule'])]

-- any defined symbol in r1 never occurs in r2
isComposableHierarchy :: TRS -> TRS -> Bool
isComposableHierarchy r1 r2 =
  isHierarchicalCombination r1 r2

composableHierarchy :: Decomposition
composableHierarchy trs =
  [(r1,r2) |
    r1 <- power trs,
    r2 <- power trs,
    not (subset r1 r2) && not (subset r2 r1),
    subset trs (r1 ++ r2),
    isHierarchicalCombination r1 r2]

decomposeCH :: Criterion -> Criterion
decomposeCH thm trs =
  run
    (proof
      "Composable Hierarchy"
      "All defined symbols in S1 never occur on S2. "
      trs)
    thm
    (composableHierarchy trs)
