-- Outside/Upside closedness
module OPC where

import Data.List
import Term
import TRS
import Rewriting
import Result
import RTermination
import qualified Positions as P
import qualified Proof as Proof

proofsh22 :: TRS -> String
proofsh22 trs = 
  Proof.proof
    "Critical Pair System with Outside Closedness"
    trs
    (unlines [
      "Critical Pair System of R without outside closed peaks is",
      "relatively terminating with respect to TRS."])

proofOO98 :: TRS -> String
proofOO98 trs =
  Proof.proof
    "Upside Parallel Closedness (Oyamaguchi and Ohta 1998)"
    trs
    "All pairs in CP(R) are upside parallel clsoed."

proofOO03 :: TRS -> String
proofOO03 trs = 
  Proof.proof
    "Outside Closedness (Oyamaguchi and Ohta 2003)"
    trs
    "All pairs in CP(R) are outside closed."

oo98 :: Criterion
oo98 trs
  | TRS.leftLinear trs && all (upsideParallelClosed trs) peaks =
      return (YES proof)
  | otherwise =
      return MAYBE
  where
    peaks = [(t,p,u) | (t,p,_,u) <- TRS.cpeak2 trs trs]
    proof = proofOO98 trs

oo03 :: Criterion
oo03 trs
  | TRS.leftLinear trs && all (outsideClosed trs) peaks =
      return (YES proof)
  | otherwise =
      return MAYBE
  where
    peaks = [(t,p,u) | (t,p,_,u) <- TRS.cpeak2 trs trs]
    proof = proofOO03 trs
  

-- t <-p-x|-e-> u
outsideClosed :: TRS -> (Term, Position, Term) -> Bool
outsideClosed trs (t,p,u) =
  elem u (rootStep trs t) ||
  elem t (u : (outsideStep trs u p))

outsideStep :: TRS -> Term -> Position -> [Term]
outsideStep trs t p =
  [ replace t u q |
     q <- functionPositions t,
     not (q P.> p),
     u <- rootStep trs (subtermAt t q) ]
  

-- t <-p-x|-e-> u
upsideParallelClosed :: TRS -> (Term, Position, Term) -> Bool
upsideParallelClosed trs (t,p,u) =
  elem u (parallelStep trs t) ||
  elem t (upsideStep trs u p)

upsideStep :: TRS -> Term -> Position -> [Term]
upsideStep trs t p =
  error (show trs ++ show t ++ show p)


-- CPS
sh22 :: RTermination -> Int -> Criterion
sh22 tt k trs
  | TRS.leftLinear trs && wcr k trs =
      cpsCriterion tt (outsideClosed trs) trs
  | otherwise = return MAYBE

type Filter = (Term, Position, Term) -> Bool

cpsCriterion :: RTermination -> Filter -> TRS -> IO Result
cpsCriterion tt f trs = do
  m <- tt ps trs
  case m of
    YES _ -> return (YES (proofsh22 trs))
    _     -> return MAYBE
  where
    ps = cps f trs

cps :: Filter -> TRS -> [(Term, Term)]
cps f r =
  [rule | (t, p, s, u) <- TRS.cpeak2 r r,
          not (f (t,p,u)),
          rule <- [(s,t),(s,u)]]

joinable :: Int -> TRS -> (Term,Term) -> Bool
joinable k trs (s,t) =
  intersect
    (at_most_k_steps k trs s)
    (at_most_k_steps k trs t) /= []

wcr :: Int -> TRS -> Bool
wcr k trs = all (joinable k trs) (cp trs)
